"""Snapshot module - provides file versioning with automatic backend selection.

When in a git repository, uses git stash for snapshots.
Otherwise, falls back to file-based snapshots in .aye/snapshots.
"""

import subprocess
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from .base import SnapshotBackend
from .file_backend import FileBasedBackend, SNAP_ROOT, LATEST_SNAP_DIR
from .git_backend import GitStashBackend

__all__ = [
    # Classes
    'SnapshotBackend',
    'FileBasedBackend',
    'GitStashBackend',
    # Constants
    'SNAP_ROOT',
    'LATEST_SNAP_DIR',
    # Public API
    'create_snapshot',
    'list_snapshots',
    'restore_snapshot',
    'apply_updates',
    'list_all_snapshots',
    'delete_snapshot',
    'prune_snapshots',
    'cleanup_snapshots',
    # Utilities
    'get_backend',
    'reset_backend',
    # Legacy helpers (for backward compatibility)
    '_get_next_ordinal',
    '_get_latest_snapshot_dir',
    '_truncate_prompt',
    '_list_all_snapshots_with_metadata',
    '_is_git_repository',
]

# ------------------------------------------------------------------
# Backend Factory
# ------------------------------------------------------------------
_backend: Optional[SnapshotBackend] = None


def _is_git_repository() -> Optional[Path]:
    """Check if current directory is inside a git repository.

    Returns:
        Path to git root if in a repo, None otherwise
    """
    try:
        result = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output=True,
            text=True,
            check=True
        )
        return Path(result.stdout.strip())
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None


def get_backend() -> SnapshotBackend:
    """Get the appropriate snapshot backend (singleton pattern)."""
    global _backend

    if _backend is None:
        git_root = _is_git_repository()
        #if git_root:
        #    _backend = GitStashBackend(git_root)
        #else:
        #    _backend = FileBasedBackend()
        _backend = FileBasedBackend()

    return _backend


def reset_backend() -> None:
    """Reset the backend singleton (useful for testing)."""
    global _backend
    _backend = None


# ------------------------------------------------------------------
# Public API (delegates to backend)
# ------------------------------------------------------------------
def create_snapshot(file_paths: List[Path], prompt: Optional[str] = None) -> str:
    """Snapshot the current contents of the given files."""
    return get_backend().create_snapshot(file_paths, prompt)


def list_snapshots(file: Optional[Path] = None) -> Union[List[str], List[Tuple[str, str]]]:
    """Return all batch-snapshot timestamps, newest first, or snapshots for a specific file."""
    return get_backend().list_snapshots(file)


def restore_snapshot(ordinal: Optional[str] = None, file_name: Optional[str] = None) -> None:
    """Restore files from a batch snapshot."""
    return get_backend().restore_snapshot(ordinal, file_name)


def apply_updates(updated_files: List[Dict[str, str]], prompt: Optional[str] = None) -> str:
    """
    1. Take a snapshot of the *current* files.
    2. Write the new contents supplied by the LLM.
    Returns the batch timestamp (useful for UI feedback).
    """
    file_paths: List[Path] = [Path(item["file_name"]) for item in updated_files if "file_name" in item and "file_content" in item]
    batch_ts = create_snapshot(file_paths, prompt)
    for item in updated_files:
        fp = Path(item["file_name"])
        fp.parent.mkdir(parents=True, exist_ok=True)
        fp.write_text(item["file_content"], encoding="utf-8")
    return batch_ts


def list_all_snapshots() -> List[Path]:
    """List all snapshot directories in chronological order (oldest first)."""
    return get_backend().list_all_snapshots()


def delete_snapshot(snapshot_dir: Path) -> None:
    """Delete a snapshot directory/entry."""
    return get_backend().delete_snapshot(snapshot_dir)


def prune_snapshots(keep_count: int = 10) -> int:
    """Delete all but the most recent N snapshots. Returns number of deleted snapshots."""
    return get_backend().prune_snapshots(keep_count)


def cleanup_snapshots(older_than_days: int = 30) -> int:
    """Delete snapshots older than N days. Returns number of deleted snapshots."""
    return get_backend().cleanup_snapshots(older_than_days)


# ------------------------------------------------------------------
# Legacy helper functions (for backward compatibility with tests)
# ------------------------------------------------------------------
def _get_next_ordinal() -> int:
    """Get the next ordinal number - delegates to file-based backend."""
    backend = get_backend()
    if isinstance(backend, FileBasedBackend):
        return backend._get_next_ordinal()
    elif isinstance(backend, GitStashBackend):
        return backend._get_next_ordinal()
    return 1


def _get_latest_snapshot_dir() -> Optional[Path]:
    """Get the latest snapshot directory by finding the one with the highest ordinal."""
    backend = get_backend()
    if isinstance(backend, FileBasedBackend):
        if not backend.snap_root.is_dir():
            return None

        snapshot_dirs = []
        for batch_dir in backend.snap_root.iterdir():
            if batch_dir.is_dir() and "_" in batch_dir.name and batch_dir.name != "latest":
                try:
                    ordinal = int(batch_dir.name.split("_")[0])
                    snapshot_dirs.append((ordinal, batch_dir))
                except ValueError:
                    continue

        if not snapshot_dirs:
            return None

        snapshot_dirs.sort(key=lambda x: x[0])
        return snapshot_dirs[-1][1]
    return None


def _truncate_prompt(prompt: Optional[str], max_length: int = 32) -> str:
    """Truncate a prompt to max_length characters."""
    backend = get_backend()
    if hasattr(backend, '_truncate_prompt'):
        return backend._truncate_prompt(prompt, max_length)
    if not prompt:
        return "no prompt".ljust(max_length)
    prompt = prompt.strip()
    if not prompt:
        return "no prompt".ljust(max_length)
    if len(prompt) <= max_length:
        return prompt.ljust(max_length)
    return prompt[:max_length] + "..."


def _list_all_snapshots_with_metadata() -> List[str]:
    """List all snapshots in descending order with file names from metadata."""
    backend = get_backend()
    if isinstance(backend, FileBasedBackend):
        return backend._list_all_snapshots_with_metadata()
    # For GitStashBackend, delegate to list_snapshots
    return backend.list_snapshots()


def driver():
    list_snapshots()


if __name__ == "__main__":
    driver()



import unittest
from unittest.mock import patch, MagicMock
import os

import aye.model.onnx_manager as onnx_manager

class TestOnnxManager(unittest.TestCase):
    def setUp(self):
        # Reset global state
        onnx_manager._status = 'NOT_CHECKED'
        # Ensure flag file doesn't exist
        if onnx_manager._model_flag_file.exists():
            onnx_manager._model_flag_file.unlink()

    def tearDown(self):
        # Clean up
        if onnx_manager._model_flag_file.exists():
            onnx_manager._model_flag_file.unlink()

    def test_get_model_status_initial(self):
        self.assertEqual(onnx_manager.get_model_status(), 'NOT_DOWNLOADED')

    @patch('chromadb.utils.embedding_functions.ONNXMiniLM_L6_V2')
    def test_get_model_status_downloading(self, mock_onnx):
        # Simulate download in progress by setting status manually
        onnx_manager._status = 'DOWNLOADING'
        status = onnx_manager.get_model_status()
        self.assertEqual(status, 'DOWNLOADING')
        # Call again to check caching
        status = onnx_manager.get_model_status()
        self.assertEqual(status, 'DOWNLOADING')

    @patch('chromadb.utils.embedding_functions.ONNXMiniLM_L6_V2')
    def test_get_model_status_ready(self, mock_onnx):
        # Create the flag file to simulate ready status
        onnx_manager._model_flag_file.parent.mkdir(parents=True, exist_ok=True)
        onnx_manager._model_flag_file.touch()
        status = onnx_manager.get_model_status()
        self.assertEqual(status, 'READY')

    @patch('chromadb.utils.embedding_functions.ONNXMiniLM_L6_V2')
    def test_get_model_status_failed(self, mock_onnx):
        # Simulate failed by setting status manually
        onnx_manager._status = 'FAILED'
        status = onnx_manager.get_model_status()
        self.assertEqual(status, 'FAILED')

    @patch('chromadb.utils.embedding_functions.ONNXMiniLM_L6_V2')
    def test_get_model_status_thread_safety(self, mock_onnx):
        # Simulate concurrent calls
        import threading
        results = []
        def call_get_status():
            results.append(onnx_manager.get_model_status())
        threads = [threading.Thread(target=call_get_status) for _ in range(5)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()
        # All should be consistent
        self.assertTrue(all(r == results[0] for r in results))

if __name__ == '__main__':
    unittest.main()


# config.py
import json
from pathlib import Path
from typing import Any, Dict

# Configuration file path
CONFIG_FILE = Path(".aye/config.json").resolve()

# Private storage – the leading underscore signals "internal".
_config: Dict[str, Any] = {}

# Default ignore patterns for file scanning
DEFAULT_IGNORE_SET = {
    'venv', 'env', 'node_modules', '__pycache__', 'dist', 'build',
    'target', 'bin', 'public', 'vendor'
}

# Shared system prompt for all LLM interactions
SYSTEM_PROMPT = (
    "You are a helpful assistant Archie, and you help users to use Aye Chat. "
    "## About Aye Chat\n\n"
    "Aye Chat is an AI-powered terminal workspace that brings AI directly into command-line workflows. "
    "It allows developers to edit files, run commands, and chat with their codebase without leaving the terminal.\n\n"
    "At the core of Aye Chat is the **optimistic workflow**: files are written directly (optimistic: LLM assumed to be right most of the time) but user has ability to restore changes with `restore` command.\n\n"
    "### Core Features:\n"
    "- **Instant undo**: `restore` or `undo` commands revert any AI changes immediately\n"
    "- **Zero config**: Automatically reads project files (respects .gitignore and .ayeignore and skips those folders and files)\n"
    "- **Real shell**: Run any command (git, pytest, vim) without leaving chat. This is direct process execution, not AI-orchestrated.\n"
    "- **Local backups**: All changes stored in `.aye/` directory\n"
    "- **RAG-powered**: Uses vector database for intelligent context retrieval\n\n"
    "### How Input is Handled (Priority Order):\n"
    "1. **Built-in Commands**: Special Aye commands like `restore`, `model`, `help`\n"
    "2. **Shell Commands**: Any valid system command (ls, git status, docker ps, etc.)\n"
    "3. **AI Prompts**: Everything else is treated as a prompt to the AI\n\n"
    "### Session & Model Control:\n"
    "- `new` - Start a fresh chat session\n"
    "- `model` - Select a different AI model\n"
    "- `verbose [on|off]` - Toggle verbose output\n"
    "- `debug [on|off]` - Toggle debug mode\n"
    "- `exit`, `quit`, Ctrl+D - Exit the chat\n"
    "- `help` - Show available commands\n\n"
    "### Reviewing & Undoing AI Changes:\n"
    "- `restore` or `undo` - Undo the last set of AI changes\n"
    "- `restore <ordinal>` - Restore to a specific snapshot (e.g., `restore 001`)\n"
    "- `restore <ordinal> <file>` - Restore a specific file from a snapshot\n"
    "- `history` - Show the history of snapshots\n"
    "- `diff <file>` - Compare current version against last snapshot\n"
    "- `diff <file> <snap1> <snap2>` - Compare two snapshots\n\n"
    "### Special Commands:\n"
    "- `with <files>: <prompt>` - Include specific files in the prompt (supports wildcards)\n"
    "  Example: `with src/*.py: refactor to use dependency injection`\n"
    "  Example: `with main.py, utils.py: explain the interaction`\n"
    "- `cd <directory>` - Change current working directory\n\n"
    "### Shell Integration:\n"
    "- Any command not recognized as built-in is executed as a shell command\n"
    "- Interactive programs work: `vim`, `nano`, `less`, `top`\n"
    "- Examples: `ls -la`, `git status`, `pytest`\n\n"
    "### Starting a Session:\n"
    "- `aye chat` - Start chat with auto-detected files\n"
    "- `aye chat --root ./src` - Specify project root\n"
    "- `aye chat --include \"*.js,*.css\"` - Manually specify file patterns\n\n"
    "### Plugin System:\n"
    "- Extensible via plugins in `~/.aye/plugins/`\n"
    "- Core plugins: shell_executor, completer, auto_detect_mask, local_model, offline_llm\n"
    "- Plugins downloaded automatically on login\n\n"
    "### Privacy & Security:\n"
    "- Respects `.gitignore` and `.ayeignore` - private files never touched\n"
    "- All backups stored locally in `.aye/` folder\n"
    "- No telemetry or usage tracking\n\n"
    "You provide clear and concise answers. Answer **directly**, give only the "
    "information the user asked for. When you are unsure, say so. You generate your responses in "
    "text-friendly format because your responses will be displayed in a terminal: use ASCII and pseudo-graphics.\n\n"
    "You follow instructions closely and respond accurately to a given prompt. You emphasize precise "
    "instruction-following and accuracy over speed of response: take your time to understand a question.\n\n"
    "Focus on accuracy in your response and follow the instructions precisely. At the same time, keep "
    "your answers brief and concise unless asked otherwise. Keep the tone professional and neutral.\n\n"
    "There may be source files appended to a user question, only use them if a question asks for help "
    "with code generation or troubleshooting; ignore them if a question is not software code related.\n\n"
    "UNDER NO CIRCUMSTANCES YOU ARE TO UPDATE SOURCE FILES UNLESS EXPLICITLY ASKED.\n\n"
    "When asked to do updates or implement features - you generate full files only as they will be "
    "inserted as is. Do not use diff notation: return only clean full files.\n\n"
    "You MUST respond with a JSON object that conforms to this schema:\n"
    '{\n'
    '    "type": "object",\n'
    '    "properties": {\n'
    '        "answer_summary": {\n'
    '            "type": "string",\n'
    '            "description": "Detailed answer to a user question"\n'
    '        },\n'
    '        "source_files": {\n'
    '            "type": "array",\n'
    '            "items": {\n'
    '                "type": "object",\n'
    '                "properties": {\n'
    '                    "file_name": {\n'
    '                        "type": "string",\n'
    '                        "description": "Name of the source file including relative path"\n'
    '                    },\n'
    '                    "file_content": {\n'
    '                        "type": "string",\n'
    '                        "description": "Full text/content of the source file"\n'
    '                    }\n'
    '                },\n'
    '                "required": ["file_name", "file_content"],\n'
    '                "additionalProperties": false\n'
    '            }\n'
    '        }\n'
    '    },\n'
    '    "required": ["answer_summary", "source_files"],\n'
    '    "additionalProperties": false\n'
    '}'
)

# Models configuration (order unchanged)
MODELS = [
    #{"id": "openai/gpt-oss-120b", "name": "OpenAI: GPT OSS 120b"},
    {"id": "x-ai/grok-code-fast-1", "name": "xAI: Grok Code Fast 1"},
    {"id": "x-ai/grok-4-fast", "name": "xAI: Grok 4 Fast"},
    #{"id": "qwen/qwen3-coder", "name": "Qwen: Qwen3 Coder"},
    #{"id": "deepseek/deepseek-chat-v3-0324", "name": "DeepSeek: DeepSeek V3 0324"},
    {"id": "google/gemini-2.0-flash-001", "name": "Google: Gemini 2.0 Flash"},
    {"id": "openai/gpt-5.1-codex-mini", "name": "OpenAI: GPT-5.1-Codex-Mini"},
    {"id": "moonshotai/kimi-k2-0905", "name": "MoonshotAI: Kimi K2 0905"},
    {"id": "google/gemini-2.5-pro", "name": "Google: Gemini 2.5 Pro"},
    {"id": "google/gemini-3-pro-preview", "name": "Google: Gemini 3 Pro Preview"},
    {"id": "anthropic/claude-sonnet-4.5", "name": "Anthropic: Claude Sonnet 4.5"},
    {"id": "openai/gpt-5.1-codex", "name": "OpenAI: GPT-5.1-Codex"},
    {"id": "openai/gpt-5.1", "name": "OpenAI: GPT-5.1"},
    {"id": "anthropic/claude-opus-4.5", "name": "Anthropic: Claude Opus 4.5"},
    
    # Offline models
    #{"id": "offline/deepseek-coder-6.7b", "name": "DeepSeek Coder 6.7B (Offline)", "type": "offline", "size_gb": 3.8},
    {"id": "offline/qwen2.5-coder-7b", "name": "Qwen2.5 Coder 7B (Offline)", "type": "offline", "size_gb": 4.7},
]

# Default model identifier – kept separate so the order of MODELS stays unchanged.
DEFAULT_MODEL_ID = "google/gemini-2.5-pro"


def load_config() -> None:
    """Load configuration from file if it exists."""
    if CONFIG_FILE.exists():
        try:
            _config.update(json.loads(CONFIG_FILE.read_text(encoding="utf-8")))
        except json.JSONDecodeError:
            pass  # Ignore invalid config files


def save_config() -> None:
    """Save configuration to file."""
    CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
    CONFIG_FILE.write_text(json.dumps(_config, indent=2), encoding="utf-8")


def get_value(key: str, default: Any = None) -> Any:
    """Return the value for *key* or *default* if the key is missing."""
    return _config.get(key, default)


def set_value(key: str, value: Any) -> None:
    """Store *value* under *key* after a simple validation."""
    if not isinstance(key, str):
        raise TypeError("Configuration key must be a string")
    # You could add more validation here (type checking, range, etc.)
    _config[key] = value
    save_config()


def delete_value(key: str) -> bool:
    """Delete a key from configuration. Returns True if key existed and was deleted."""
    if key in _config:
        del _config[key]
        save_config()
        return True
    return False


def list_config() -> Dict[str, Any]:
    """Return a copy of the current configuration."""
    return _config.copy()


def driver() -> None:
    """Simple driver to demonstrate loading and listing the configuration.

    When executed directly (`python config.py`), this function loads the
    configuration from ``.aye/config.json`` (if present) and prints the current
    settings as pretty‑printed JSON to stdout.
    """
    load_config()
    cfg = list_config()
    # Print the configuration in a readable format; an empty config results in
    # an empty JSON object.
    print(json.dumps(cfg, indent=2))


if __name__ == "__main__":
    driver()



# Test suite for aye.model.api module
import os
import json
import time
from unittest import TestCase
from unittest.mock import patch, MagicMock
import httpx
import aye.model.api as api
from aye.model.auth import get_token


class TestModelApi(TestCase):
    def setUp(self):
        self.base_url = "https://api.ayechat.ai"
        self.token = "fake_token"
        os.environ["AYE_TOKEN"] = self.token  # Set env for testing

    def tearDown(self):
        if "AYE_TOKEN" in os.environ:
            del os.environ["AYE_TOKEN"]

    @patch('aye.model.api.get_token')
    def test_auth_headers(self, mock_get_token):
        mock_get_token.return_value = self.token
        headers = api._auth_headers()
        self.assertEqual(headers, {"Authorization": f"Bearer {self.token}"}) 

    @patch('aye.model.api.get_token')
    def test_auth_headers_no_token(self, mock_get_token):
        mock_get_token.return_value = None
        with self.assertRaises(RuntimeError) as cm:
            api._auth_headers()
        self.assertIn("No auth token", str(cm.exception))

    def test_check_response_success(self):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = {"data": "ok"}
        mock_resp.text = "ok"
        mock_resp.raise_for_status.return_value = None
        result = api._check_response(mock_resp)
        self.assertEqual(result, {"data": "ok"})

    def test_check_response_error_status(self):
        mock_resp = MagicMock()
        mock_resp.status_code = 400
        mock_resp.json.return_value = {"error": "Bad request"}
        mock_resp.text = "Bad request"
        mock_resp.raise_for_status.side_effect = httpx.HTTPStatusError("error", request=None, response=mock_resp)
        with self.assertRaises(Exception) as cm:
            api._check_response(mock_resp)
        self.assertIn("Bad request", str(cm.exception))

    def test_check_response_error_status_non_json(self):
        mock_resp = MagicMock()
        mock_resp.status_code = 400
        mock_resp.json.side_effect = json.JSONDecodeError("", "", 0)
        mock_resp.text = "Raw error text"
        mock_resp.raise_for_status.side_effect = httpx.HTTPStatusError("error", request=None, response=mock_resp)
        with self.assertRaises(Exception) as cm:
            api._check_response(mock_resp)
        self.assertIn("Raw error text", str(cm.exception))

    def test_check_response_json_error(self):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.raise_for_status.return_value = None
        mock_resp.json.return_value = {"error": "Server error"}
        mock_resp.text = "Server error"
        with self.assertRaises(Exception) as cm:
            api._check_response(mock_resp)
        self.assertIn("Server error", str(cm.exception))

    def test_check_response_non_json(self):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.raise_for_status.return_value = None
        mock_resp.json.side_effect = json.JSONDecodeError("", "", 0)
        mock_resp.text = "plain text"
        result = api._check_response(mock_resp)
        self.assertEqual(result, {})

    @patch('aye.model.api.time')
    @patch('httpx.get')
    @patch('httpx.Client')
    @patch('aye.model.api._check_response')
    @patch('aye.model.api._auth_headers')
    def test_cli_invoke_polling_success(self, mock_headers, mock_check, mock_client, mock_get, mock_time):
        mock_headers.return_value = {"Auth": "fake"}
        mock_post_resp = MagicMock()
        mock_post_resp.json.return_value = {"response_url": "https://fake.url"}
        mock_client.return_value.__enter__.return_value.post.return_value = mock_post_resp
        mock_check.return_value = {"response_url": "https://fake.url"}

        # Mock polling: first 404, then 200 with final data
        mock_time.sleep.return_value = None
        mock_time.time.side_effect = [0, 2, 4]
        mock_get.side_effect = [
            MagicMock(status_code=404),
            MagicMock(status_code=200, json=lambda: {"final": "response"})
        ]

        result = api.cli_invoke(message="test", dry_run=False)
        self.assertEqual(result, {"final": "response"})
        self.assertEqual(mock_get.call_count, 2)
        mock_get.assert_called_with("https://fake.url", timeout=api.TIMEOUT)

    @patch('aye.model.api.time')
    @patch('httpx.get')
    @patch('httpx.Client')
    @patch('aye.model.api._check_response')
    @patch('aye.model.api._auth_headers')
    def test_cli_invoke_polling_json_decode_error(self, mock_headers, mock_check, mock_client, mock_get, mock_time):
        mock_headers.return_value = {"Auth": "fake"}
        mock_check.return_value = {"response_url": "https://fake.url"}
        mock_time.time.side_effect = [0, 2]
        mock_get.return_value = MagicMock(status_code=200)
        mock_get.return_value.json.side_effect = json.JSONDecodeError("err", "doc", 0)

        with self.assertRaises(json.JSONDecodeError):
            api.cli_invoke(message="test")

    @patch('aye.model.api.time')
    @patch('httpx.get')
    @patch('httpx.Client')
    @patch('aye.model.api._check_response')
    @patch('aye.model.api._auth_headers')
    def test_cli_invoke_polling_request_error(self, mock_headers, mock_check, mock_client, mock_get, mock_time):
        mock_headers.return_value = {"Auth": "fake"}
        mock_check.return_value = {"response_url": "https://fake.url"}
        mock_time.time.side_effect = [0, 2, 4]
        mock_get.side_effect = [
            httpx.RequestError("network error"),
            MagicMock(status_code=200, json=lambda: {"final": "response"})
        ]

        result = api.cli_invoke(message="test")
        self.assertEqual(result, {"final": "response"})
        self.assertEqual(mock_get.call_count, 2)

    @patch('aye.model.api.time')
    @patch('httpx.get')
    @patch('httpx.Client')
    @patch('aye.model.api._check_response')
    @patch('aye.model.api._auth_headers')
    def test_cli_invoke_timeout(self, mock_headers, mock_check, mock_client, mock_get, mock_time):
        mock_headers.return_value = {"Auth": "fake"}
        mock_post_resp = MagicMock()
        mock_post_resp.json.return_value = {"response_url": "https://fake.url"}
        mock_client.return_value.__enter__.return_value.post.return_value = mock_post_resp
        mock_check.return_value = {"response_url": "https://fake.url"}

        mock_time.sleep.return_value = None
        deadline = 120
        timestamps = list(range(0, deadline, 2)) + [deadline + 1]
        mock_time.time.side_effect = timestamps
        mock_get.return_value = MagicMock(status_code=404)

        with self.assertRaises(TimeoutError):
            api.cli_invoke(message="test", dry_run=False, poll_timeout=deadline)

    @patch('aye.model.api._auth_headers')
    @patch('httpx.Client')
    def test_fetch_plugin_manifest_success(self, mock_client, mock_headers):
        mock_headers.return_value = {"Auth": "fake"}
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = {"plugins": "data"}
        mock_client.return_value.__enter__.return_value.post.return_value = mock_resp

        result = api.fetch_plugin_manifest(dry_run=True)
        self.assertEqual(result, {"plugins": "data"})

    @patch('aye.model.api._auth_headers')
    @patch('httpx.Client')
    def test_fetch_plugin_manifest_error(self, mock_client, mock_headers):
        mock_headers.return_value = {"Auth": "fake"}
        mock_resp = MagicMock()
        mock_resp.status_code = 500
        mock_resp.json.return_value = {"error": "Server error"}
        mock_resp.text = '{"error": "Server error"}'
        mock_resp.raise_for_status.side_effect = httpx.HTTPStatusError("error", request=None, response=mock_resp)
        mock_client.return_value.__enter__.return_value.post.return_value = mock_resp

        with self.assertRaises(Exception) as cm:
            api.fetch_plugin_manifest(dry_run=True)
        self.assertIn("Server error", str(cm.exception))

    @patch('aye.model.api._auth_headers')
    @patch('httpx.Client')
    def test_fetch_server_time_success(self, mock_client, mock_headers):
        mock_headers.return_value = {"Auth": "fake"}
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.ok = True
        mock_resp.json.return_value = {"timestamp": 1234567890}
        mock_client.return_value.__enter__.return_value.get.return_value = mock_resp

        result = api.fetch_server_time(dry_run=True)
        self.assertEqual(result, 1234567890)

    @patch('aye.model.api._auth_headers')
    @patch('httpx.Client')
    def test_fetch_server_time_error(self, mock_client, mock_headers):
        mock_headers.return_value = {"Auth": "fake"}
        mock_resp = MagicMock()
        mock_resp.status_code = 500
        mock_resp.ok = False
        mock_resp.json.return_value = {"error": "Server error"}
        mock_resp.text = '{"error": "Server error"}'
        mock_resp.raise_for_status.side_effect = httpx.HTTPStatusError("error", request=None, response=mock_resp)
        mock_client.return_value.__enter__.return_value.get.return_value = mock_resp

        with self.assertRaises(Exception) as cm:
            api.fetch_server_time(dry_run=True)
        self.assertIn("Server error", str(cm.exception))

    @patch('aye.model.api._auth_headers')
    @patch('httpx.Client')
    def test_send_feedback_success(self, mock_client, mock_headers):
        mock_headers.return_value = {"Auth": "fake"}
        mock_post_resp = MagicMock()
        mock_post_resp.status_code = 200
        mock_client.return_value.__enter__.return_value.post.return_value = mock_post_resp

        api.send_feedback("great tool!", chat_id=123)

        mock_client.return_value.__enter__.return_value.post.assert_called_once()
        call_args = mock_client.return_value.__enter__.return_value.post.call_args
        #self.assertEqual(call_args.args[0], f"{self.base_url}/feedback")

        self.assertTrue('ayechat.ai/feedback' in call_args.args[0])
        self.assertEqual(call_args.kwargs['json'], {"feedback": "great tool!", "chat_id": 123})

    @patch('aye.model.api._auth_headers')
    @patch('httpx.Client')
    def test_send_feedback_error_ignored(self, mock_client, mock_headers):
        mock_headers.return_value = {"Auth": "fake"}
        mock_client.return_value.__enter__.return_value.post.side_effect = httpx.RequestError("network error")

        # Should not raise an exception
        api.send_feedback("this will fail silently", chat_id=123)
        mock_client.return_value.__enter__.return_value.post.assert_called_once()

    @patch('builtins.print')
    @patch('aye.model.api.get_user_config', return_value='on')
    def test_debug_mode_prints(self, mock_get_config, mock_print):
        # Test cli_invoke
        with patch('httpx.Client') as mock_client, \
             patch('httpx.get') as mock_get, \
             patch('aye.model.api._auth_headers'):
            mock_post_resp = MagicMock()
            mock_get_resp = MagicMock()

            mock_post_resp.status_code = 200
            mock_post_resp.json.return_value = {"response_url": "https://testurl"}
            mock_client.return_value.__enter__.return_value.post.return_value = mock_post_resp

            mock_get_resp.status_code = 200
            mock_get_resp.json.return_value = { "answer_summary": "Test response", "source_files": []}
            mock_get.return_value = mock_get_resp
            api.cli_invoke(message="test")
            self.assertIn("[DEBUG] Sending request to", str(mock_print.call_args_list[0][0][0]))

        # Test fetch_plugin_manifest
        with patch('httpx.Client') as mock_client, patch('aye.model.api._auth_headers'):
            mock_post_resp = MagicMock()
            mock_post_resp.status_code = 200
            mock_post_resp.json.return_value = {"plugins": "data"}
            mock_client.return_value.__enter__.return_value.post.return_value = mock_post_resp
            api.fetch_plugin_manifest()
            # Find the call after the first three (sending, payload, headers)
            debug_calls = [str(call[0][0]) for call in mock_print.call_args_list if "[DEBUG] Sending request to" in str(call[0][0])]
            self.assertIn("[DEBUG] Sending request to", debug_calls[-1])

        # Test fetch_server_time
        with patch('httpx.Client') as mock_client, patch('aye.model.api._auth_headers'):
            mock_get_resp = MagicMock()
            mock_get_resp.status_code = 200
            mock_get_resp.ok = True
            mock_get_resp.json.return_value = {"timestamp": 123}
            mock_client.return_value.__enter__.return_value.get.return_value = mock_get_resp
            api.fetch_server_time()
            debug_calls = [str(call[0][0]) for call in mock_print.call_args_list if "[DEBUG] Sending request to" in str(call[0][0])]
            self.assertIn("[DEBUG] Sending request to", debug_calls[-1])

        # Test send_feedback
        with patch('httpx.Client') as mock_client, patch('aye.model.api._auth_headers'):
            mock_post_resp = MagicMock()
            mock_post_resp.status_code = 200
            mock_client.return_value.__enter__.return_value.post.return_value = mock_post_resp
            api.send_feedback("feedback")
            debug_calls = [str(call[0][0]) for call in mock_print.call_args_list if "[DEBUG] Sending request to" in str(call[0][0])]
            self.assertIn("[DEBUG] Sending request to", debug_calls[-1])

        # Test send_feedback error in debug
        with patch('httpx.Client') as mock_client, patch('aye.model.api._auth_headers'):
            mock_client.return_value.__enter__.return_value.post.side_effect = Exception("send error")
            api.send_feedback("feedback")
            self.assertIn("[DEBUG] Error sending feedback: send error", str(mock_print.call_args_list[-1][0][0]))



import os
import pytest
from pathlib import Path
from unittest.mock import Mock, patch, MagicMock, mock_open
from prompt_toolkit import PromptSession
from rich.console import Console

from aye.controller.command_handlers import (
    handle_cd_command,
    handle_model_command,
    handle_verbose_command,
    handle_debug_command,
    _expand_file_patterns,
    handle_with_command
)


class TestHandleCdCommand:
    """Tests for handle_cd_command function."""

    def test_cd_to_home_when_no_target_provided(self, tmp_path):
        """Test cd with no arguments changes to home directory."""
        conf = Mock()
        conf.root = tmp_path
        tokens = ["cd"]
        
        with patch('os.chdir') as mock_chdir:
            result = handle_cd_command(tokens, conf)
            
            assert result is True
            mock_chdir.assert_called_once_with(str(Path.home()))
            assert conf.root == Path.cwd()

    def test_cd_to_specific_directory(self, tmp_path):
        """Test cd to a specific directory."""
        conf = Mock()
        conf.root = tmp_path
        target_dir = tmp_path / "subdir"
        target_dir.mkdir()
        tokens = ["cd", str(target_dir)]
        
        result = handle_cd_command(tokens, conf)
        
        assert result is True
        assert conf.root == Path.cwd()

    def test_cd_with_spaces_in_path(self, tmp_path):
        """Test cd with directory name containing spaces."""
        conf = Mock()
        conf.root = tmp_path
        target_dir = tmp_path / "dir with spaces"
        target_dir.mkdir()
        tokens = ["cd", "dir", "with", "spaces"]
        
        with patch('os.chdir') as mock_chdir:
            result = handle_cd_command(tokens, conf)
            
            assert result is True
            mock_chdir.assert_called_once_with("dir with spaces")

    def test_cd_to_nonexistent_directory(self, tmp_path):
        """Test cd to a directory that doesn't exist."""
        conf = Mock()
        conf.root = tmp_path
        tokens = ["cd", "/nonexistent/path"]
        
        with patch('aye.controller.command_handlers.print_error') as mock_print_error:
            result = handle_cd_command(tokens, conf)
            
            assert result is False
            mock_print_error.assert_called_once()


class TestHandleModelCommand:
    """Tests for handle_model_command function."""

    @pytest.fixture
    def mock_models(self):
        return [
            {"id": "model-1", "name": "Model One", "type": "online"},
            {"id": "model-2", "name": "Model Two", "type": "offline", "size_gb": 5},
            {"id": "model-3", "name": "Model Three", "type": "online"}
        ]

    @pytest.fixture
    def mock_conf(self):
        conf = Mock()
        conf.selected_model = "model-1"
        conf.plugin_manager = Mock()
        conf.plugin_manager.handle_command = Mock(return_value={"success": True})
        return conf

    def test_select_model_by_number(self, mock_models, mock_conf):
        """Test selecting a model by number."""
        tokens = ["model", "2"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_model_command(None, mock_models, mock_conf, tokens)
            
            assert mock_conf.selected_model == "model-2"
            mock_set_config.assert_called_once_with("selected_model", "model-2")

    def test_select_offline_model_triggers_download(self, mock_models, mock_conf):
        """Test selecting an offline model triggers download."""
        tokens = ["model", "2"]
        
        with patch('aye.controller.command_handlers.set_user_config'):
            handle_model_command(None, mock_models, mock_conf, tokens)
            
            mock_conf.plugin_manager.handle_command.assert_called_once_with(
                "download_offline_model",
                {
                    "model_id": "model-2",
                    "model_name": "Model Two",
                    "size_gb": 5
                }
            )

    def test_select_offline_model_download_fails(self, mock_models, mock_conf):
        """Test handling failed offline model download."""
        mock_conf.plugin_manager.handle_command = Mock(
            return_value={"success": False, "error": "Download failed"}
        )
        tokens = ["model", "2"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_model_command(None, mock_models, mock_conf, tokens)
            
            # Model should not be selected if download fails
            mock_set_config.assert_not_called()

    def test_select_invalid_model_number(self, mock_models, mock_conf):
        """Test selecting an invalid model number."""
        tokens = ["model", "99"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_model_command(None, mock_models, mock_conf, tokens)
            
            mock_set_config.assert_not_called()

    def test_select_model_with_invalid_input(self, mock_models, mock_conf):
        """Test selecting a model with non-numeric input."""
        tokens = ["model", "invalid"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_model_command(None, mock_models, mock_conf, tokens)
            
            mock_set_config.assert_not_called()

    def test_list_models_without_session(self, mock_models, mock_conf):
        """Test listing models without a session."""
        tokens = ["model"]
        
        # Should not raise an exception
        handle_model_command(None, mock_models, mock_conf, tokens)

    def test_interactive_model_selection(self, mock_models, mock_conf):
        """Test interactive model selection with session."""
        mock_session = Mock(spec=PromptSession)
        mock_session.prompt = Mock(return_value="3")
        tokens = ["model"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_model_command(mock_session, mock_models, mock_conf, tokens)
            
            assert mock_conf.selected_model == "model-3"
            mock_set_config.assert_called_once_with("selected_model", "model-3")

    def test_interactive_model_selection_cancelled(self, mock_models, mock_conf):
        """Test interactive model selection when user presses Enter."""
        mock_session = Mock(spec=PromptSession)
        mock_session.prompt = Mock(return_value="")
        tokens = ["model"]
        original_model = mock_conf.selected_model
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_model_command(mock_session, mock_models, mock_conf, tokens)
            
            assert mock_conf.selected_model == original_model
            mock_set_config.assert_not_called()


class TestHandleVerboseCommand:
    """Tests for handle_verbose_command function."""

    def test_set_verbose_on(self):
        """Test setting verbose mode on."""
        tokens = ["verbose", "on"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_verbose_command(tokens)
            
            mock_set_config.assert_called_once_with("verbose", "on")

    def test_set_verbose_off(self):
        """Test setting verbose mode off."""
        tokens = ["verbose", "off"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_verbose_command(tokens)
            
            mock_set_config.assert_called_once_with("verbose", "off")

    def test_set_verbose_invalid_value(self):
        """Test setting verbose mode with invalid value."""
        tokens = ["verbose", "invalid"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_verbose_command(tokens)
            
            mock_set_config.assert_not_called()

    def test_get_verbose_status(self):
        """Test getting current verbose status."""
        tokens = ["verbose"]
        
        with patch('aye.controller.command_handlers.get_user_config', return_value="on"):
            handle_verbose_command(tokens)


class TestHandleDebugCommand:
    """Tests for handle_debug_command function."""

    def test_set_debug_on(self):
        """Test setting debug mode on."""
        tokens = ["debug", "on"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_debug_command(tokens)
            
            mock_set_config.assert_called_once_with("debug", "on")

    def test_set_debug_off(self):
        """Test setting debug mode off."""
        tokens = ["debug", "off"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_debug_command(tokens)
            
            mock_set_config.assert_called_once_with("debug", "off")

    def test_set_debug_invalid_value(self):
        """Test setting debug mode with invalid value."""
        tokens = ["debug", "invalid"]
        
        with patch('aye.controller.command_handlers.set_user_config') as mock_set_config:
            handle_debug_command(tokens)
            
            mock_set_config.assert_not_called()

    def test_get_debug_status(self):
        """Test getting current debug status."""
        tokens = ["debug"]
        
        with patch('aye.controller.command_handlers.get_user_config', return_value="off"):
            handle_debug_command(tokens)


class TestExpandFilePatterns:
    """Tests for _expand_file_patterns function."""

    def test_expand_single_file(self, tmp_path):
        """Test expanding a single file pattern."""
        conf = Mock()
        conf.root = tmp_path
        test_file = tmp_path / "test.py"
        test_file.write_text("content")
        
        result = _expand_file_patterns(["test.py"], conf)
        
        assert result == ["test.py"]

    def test_expand_wildcard_pattern(self, tmp_path):
        """Test expanding wildcard patterns."""
        conf = Mock()
        conf.root = tmp_path
        (tmp_path / "file1.py").write_text("content")
        (tmp_path / "file2.py").write_text("content")
        (tmp_path / "file.txt").write_text("content")
        
        result = _expand_file_patterns(["*.py"], conf)
        
        assert len(result) == 2
        assert "file1.py" in result
        assert "file2.py" in result

    def test_expand_nested_wildcard(self, tmp_path):
        """Test expanding nested wildcard patterns."""
        conf = Mock()
        conf.root = tmp_path
        subdir = tmp_path / "src"
        subdir.mkdir()
        (subdir / "main.py").write_text("content")
        
        result = _expand_file_patterns(["src/*.py"], conf)
        
        assert len(result) == 1
        assert "src/main.py" in result or "src\\main.py" in result

    def test_expand_multiple_patterns(self, tmp_path):
        """Test expanding multiple patterns."""
        conf = Mock()
        conf.root = tmp_path
        (tmp_path / "file.py").write_text("content")
        (tmp_path / "file.txt").write_text("content")
        
        result = _expand_file_patterns(["*.py", "*.txt"], conf)
        
        assert len(result) == 2

    def test_expand_empty_pattern(self, tmp_path):
        """Test expanding empty pattern."""
        conf = Mock()
        conf.root = tmp_path
        
        result = _expand_file_patterns([""], conf)
        
        assert result == []

    def test_expand_nonexistent_pattern(self, tmp_path):
        """Test expanding pattern with no matches."""
        conf = Mock()
        conf.root = tmp_path
        
        result = _expand_file_patterns(["*.nonexistent"], conf)
        
        assert result == []

    def test_expand_directory_not_included(self, tmp_path):
        """Test that directories are not included in expansion."""
        conf = Mock()
        conf.root = tmp_path
        (tmp_path / "dir").mkdir()
        (tmp_path / "file.py").write_text("content")
        
        result = _expand_file_patterns(["*"], conf)
        
        assert "file.py" in result
        assert "dir" not in result


class TestHandleWithCommand:
    """Tests for handle_with_command function."""

    @pytest.fixture
    def mock_conf(self, tmp_path):
        conf = Mock()
        conf.root = tmp_path
        conf.verbose = False
        conf.plugin_manager = Mock()
        return conf

    @pytest.fixture
    def mock_console(self):
        return Mock(spec=Console)

    def test_with_single_file(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with a single file."""
        test_file = tmp_path / "test.py"
        test_file.write_text("print('hello')")
        
        prompt = "with test.py: explain this code"
        chat_id = 1
        chat_id_file = tmp_path / "chat_id"
        
        with patch('aye.controller.command_handlers.invoke_llm') as mock_invoke, \
             patch('aye.controller.command_handlers.process_llm_response') as mock_process:
            mock_invoke.return_value = Mock(chat_id=2)
            mock_process.return_value = 2
            
            result = handle_with_command(prompt, mock_conf, mock_console, chat_id, chat_id_file)
            
            assert result == 2
            mock_invoke.assert_called_once()
            call_kwargs = mock_invoke.call_args[1]
            assert "test.py" in call_kwargs["explicit_source_files"]

    def test_with_multiple_files(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with multiple files."""
        (tmp_path / "file1.py").write_text("content1")
        (tmp_path / "file2.py").write_text("content2")
        
        prompt = "with file1.py, file2.py: analyze these files"
        
        with patch('aye.controller.command_handlers.invoke_llm') as mock_invoke, \
             patch('aye.controller.command_handlers.process_llm_response'):
            mock_invoke.return_value = Mock(chat_id=2)
            
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
            
            call_kwargs = mock_invoke.call_args[1]
            assert "file1.py" in call_kwargs["explicit_source_files"]
            assert "file2.py" in call_kwargs["explicit_source_files"]

    def test_with_wildcard_pattern(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with wildcard pattern."""
        (tmp_path / "file1.py").write_text("content1")
        (tmp_path / "file2.py").write_text("content2")
        (tmp_path / "file.txt").write_text("content")
        
        prompt = "with *.py: analyze python files"
        
        with patch('aye.controller.command_handlers.invoke_llm') as mock_invoke, \
             patch('aye.controller.command_handlers.process_llm_response'):
            mock_invoke.return_value = Mock(chat_id=2)
            
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
            
            call_kwargs = mock_invoke.call_args[1]
            assert len(call_kwargs["explicit_source_files"]) == 2

    def test_with_empty_file_list(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with empty file list."""
        prompt = "with : some prompt"
        
        result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
        
        assert result is None

    def test_with_empty_prompt(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with empty prompt after colon."""
        prompt = "with test.py:"
        
        result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
        
        assert result is None

    def test_with_nonexistent_file(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with nonexistent file."""
        prompt = "with nonexistent.py: explain this"
        
        result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
        
        assert result is None

    def test_with_unreadable_file(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command with unreadable file."""
        test_file = tmp_path / "test.py"
        test_file.write_text("content")
        
        prompt = "with test.py: explain this"
        
        with patch('pathlib.Path.read_text', side_effect=PermissionError("Access denied")):
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
            
            assert result is None

    def test_with_partial_file_failure(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command when some files fail to read."""
        (tmp_path / "file1.py").write_text("content1")
        (tmp_path / "file2.py").write_text("content2")
        
        prompt = "with file1.py, file2.py: analyze"
        
        def mock_read_text(encoding=None):
            if "file1" in str(mock_read_text.call_count):
                raise IOError("Cannot read")
            return "content2"
        
        with patch('aye.controller.command_handlers.invoke_llm') as mock_invoke, \
             patch('aye.controller.command_handlers.process_llm_response'):
            mock_invoke.return_value = Mock(chat_id=2)
            
            # Should continue with readable files
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")

    def test_with_verbose_mode(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command in verbose mode shows file list."""
        mock_conf.verbose = True
        (tmp_path / "test.py").write_text("content")
        
        prompt = "with test.py: explain"
        
        with patch('aye.controller.command_handlers.invoke_llm') as mock_invoke, \
             patch('aye.controller.command_handlers.process_llm_response'):
            mock_invoke.return_value = Mock(chat_id=2)
            
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")

    def test_with_exception_handling(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command handles unexpected exceptions."""
        prompt = "with test.py: explain"
        
        with patch('aye.controller.command_handlers._expand_file_patterns', side_effect=Exception("Unexpected error")), \
             patch('aye.controller.command_handlers.handle_llm_error') as mock_error:
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
            
            assert result is None
            mock_error.assert_called_once()

    def test_with_no_llm_response(self, mock_conf, mock_console, tmp_path):
        """Test 'with' command when LLM returns no response."""
        (tmp_path / "test.py").write_text("content")
        prompt = "with test.py: explain"
        
        with patch('aye.controller.command_handlers.invoke_llm', return_value=None):
            result = handle_with_command(prompt, mock_conf, mock_console, 1, tmp_path / "chat_id")
            
            assert result is None



"""State management classes for IndexManager.

This module contains:
- IndexConfig: Configuration dataclass
- SafeState: Thread-safe state container
- IndexingState: All indexing state in one place
- ProgressTracker: Progress display abstraction
- InitializationCoordinator: Initialization logic
- ErrorHandler: Centralized error handling
"""

import threading
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Any, List, Optional

from rich import print as rprint


# =============================================================================
# Configuration
# =============================================================================

@dataclass
class IndexConfig:
    """Configuration for IndexManager."""
    root_path: Path
    file_mask: str
    verbose: bool = False
    debug: bool = False
    save_interval: int = 20
    max_workers: int = 2
    
    @classmethod
    def from_params(
        cls,
        root_path: Path,
        file_mask: str,
        verbose: bool = False,
        debug: bool = False
    ) -> 'IndexConfig':
        """Create config from individual parameters (backward compatibility)."""
        from .index_manager_utils import MAX_WORKERS
        return cls(
            root_path=root_path,
            file_mask=file_mask,
            verbose=verbose,
            debug=debug,
            max_workers=MAX_WORKERS
        )
    
    @property
    def index_dir(self) -> Path:
        """Get the index directory path."""
        return self.root_path / ".aye"
    
    @property
    def hash_index_path(self) -> Path:
        """Get the hash index file path."""
        return self.index_dir / "file_index.json"


# =============================================================================
# Thread-Safe State Container
# =============================================================================

class SafeState:
    """Thread-safe state container with simple get/update interface."""
    
    def __init__(self):
        self._lock = threading.Lock()
        self._data: Dict[str, Any] = {}
    
    def update(self, key: str, value: Any) -> None:
        """Update a single value."""
        with self._lock:
            self._data[key] = value
    
    def get(self, key: str, default: Any = None) -> Any:
        """Get a value."""
        with self._lock:
            return self._data.get(key, default)
    
    def update_many(self, updates: Dict[str, Any]) -> None:
        """Update multiple values atomically."""
        with self._lock:
            self._data.update(updates)
    
    def get_many(self, keys: List[str]) -> Dict[str, Any]:
        """Get multiple values atomically."""
        with self._lock:
            return {k: self._data.get(k) for k in keys}
    
    def increment(self, key: str, amount: int = 1) -> int:
        """Increment a numeric value and return the new value."""
        with self._lock:
            current = self._data.get(key, 0)
            new_value = current + amount
            self._data[key] = new_value
            return new_value


# =============================================================================
# Indexing State
# =============================================================================

@dataclass
class IndexingState:
    """
    Consolidated state for all indexing operations.
    
    Replaces multiple individual instance variables with a single state object.
    """
    # Status flags
    is_indexing: bool = False
    is_refining: bool = False
    is_discovering: bool = False
    shutdown_requested: bool = False
    
    # Progress counters
    coarse_total: int = 0
    coarse_processed: int = 0
    refine_total: int = 0
    refine_processed: int = 0
    discovery_total: int = 0
    discovery_processed: int = 0
    
    # Generation counter for invalidating old runs
    generation: int = 0
    
    # Work queues
    files_to_coarse_index: List[str] = field(default_factory=list)
    files_to_refine: List[str] = field(default_factory=list)
    
    # Index data
    target_index: Dict[str, Any] = field(default_factory=dict)
    current_index_on_disk: Dict[str, Any] = field(default_factory=dict)
    
    def reset_coarse_progress(self, total: int) -> None:
        """Reset coarse indexing progress."""
        self.coarse_total = total
        self.coarse_processed = 0
    
    def reset_refine_progress(self, total: int) -> None:
        """Reset refinement progress."""
        self.refine_total = total
        self.refine_processed = 0
    
    def reset_discovery_progress(self) -> None:
        """Reset discovery progress."""
        self.discovery_total = 0
        self.discovery_processed = 0
    
    def increment_generation(self) -> int:
        """Increment and return the new generation."""
        self.generation += 1
        return self.generation
    
    def has_work(self) -> bool:
        """Check if there's indexing work to do."""
        return bool(self.files_to_coarse_index or self.files_to_refine)
    
    def is_active(self) -> bool:
        """Check if any background work is in progress."""
        return self.is_indexing or self.is_refining or self.is_discovering
    
    def clear_work_queues(self) -> None:
        """Clear all work queues."""
        self.files_to_coarse_index = []
        self.files_to_refine = []
        self.target_index = {}


# =============================================================================
# Progress Tracker
# =============================================================================

class ProgressTracker:
    """
    Thread-safe progress tracking with display formatting.
    
    Tracks progress for three phases: discovery, coarse, refine.
    """
    
    def __init__(self):
        self._lock = threading.Lock()
        self._phases: Dict[str, Dict[str, int]] = {
            'discovery': {'processed': 0, 'total': 0},
            'coarse': {'processed': 0, 'total': 0},
            'refine': {'processed': 0, 'total': 0}
        }
        self._active_phase: Optional[str] = None
    
    def set_active(self, phase: Optional[str]) -> None:
        """Set the currently active phase."""
        with self._lock:
            self._active_phase = phase
    
    def set_total(self, phase: str, total: int) -> None:
        """Set the total count for a phase."""
        with self._lock:
            self._phases[phase]['total'] = total
            self._phases[phase]['processed'] = 0
    
    def increment(self, phase: str) -> int:
        """Increment processed count and return new value."""
        with self._lock:
            self._phases[phase]['processed'] += 1
            return self._phases[phase]['processed']
    
    def get_progress(self, phase: str) -> tuple:
        """Get (processed, total) for a phase."""
        with self._lock:
            p = self._phases[phase]
            return p['processed'], p['total']
    
    def get_display(self) -> str:
        """Get formatted progress display string."""
        acquired = self._lock.acquire(timeout=0.01)
        if not acquired:
            return "indexing..."
        try:
            phase = self._active_phase
            if phase is None:
                return ""
            
            p = self._phases[phase]
            processed, total = p['processed'], p['total']
            
            if phase == 'discovery':
                if total > 0:
                    return f"discovering files {processed}/{total}"
                return "discovering files..."
            elif phase == 'coarse':
                return f"indexing {processed}/{total}"
            elif phase == 'refine':
                return f"refining {processed}/{total}"
            return ""
        finally:
            self._lock.release()
    
    def is_active(self) -> bool:
        """Check if any phase is active."""
        with self._lock:
            return self._active_phase is not None


# =============================================================================
# Initialization Coordinator
# =============================================================================

class InitializationCoordinator:
    """
    Coordinates vector DB initialization.
    
    Handles initialization state, locking, and retry logic.
    """
    
    def __init__(self, config: IndexConfig):
        self.config = config
        self.collection: Optional[Any] = None
        self._is_initialized = False
        self._in_progress = False
        self._lock = threading.Lock()
    
    @property
    def is_initialized(self) -> bool:
        """Check if initialization is complete."""
        return self._is_initialized
    
    @property
    def in_progress(self) -> bool:
        """Check if initialization is in progress."""
        return self._in_progress
    
    @property
    def is_ready(self) -> bool:
        """Check if the collection is ready for use."""
        return self._is_initialized and self.collection is not None
    
    def initialize(self, blocking: bool = True) -> bool:
        """
        Initialize the ChromaDB collection.
        
        Args:
            blocking: If True, wait for lock. If False, return immediately
                      if lock is held.
                      
        Returns:
            True on success or if already initialized.
        """
        from aye.model import vector_db, onnx_manager
        
        # Fast path: already initialized
        if self._is_initialized:
            return self.collection is not None
        
        # Try to acquire lock
        if blocking:
            acquired = self._lock.acquire(timeout=0.1)
        else:
            acquired = self._lock.acquire(blocking=False)
        
        if not acquired:
            return self._is_initialized and self.collection is not None
        
        try:
            if self._is_initialized:
                return self.collection is not None
            
            self._in_progress = True
            model_status = onnx_manager.get_model_status()
            
            if model_status == "READY":
                return self._do_initialize()
            elif model_status == "FAILED":
                self._is_initialized = True
                self.collection = None
                return False
            
            return False
        finally:
            self._in_progress = False
            self._lock.release()
    
    def _do_initialize(self) -> bool:
        """Perform the actual initialization."""
        from aye.model import vector_db
        
        try:
            self.collection = vector_db.initialize_index(self.config.root_path)
            self._is_initialized = True
            if self.config.debug:
                rprint("[bold cyan]Code lookup is now active.[/]")
            return True
        except Exception as e:
            rprint(f"[red]Failed to initialize local code search: {e}[/red]")
            self._is_initialized = True
            self.collection = None
            return False


# =============================================================================
# Error Handler
# =============================================================================

class ErrorHandler:
    """
    Centralized error handling with context.
    
    Respects verbose/debug settings for output control.
    """
    
    def __init__(self, verbose: bool = False, debug: bool = False):
        self.verbose = verbose
        self.debug = debug
    
    def handle(self, error: Exception, context: str = "") -> None:
        """Handle an error with optional context."""
        if self.debug:
            if context:
                rprint(f"[red]Error in {context}: {error}[/red]")
            else:
                rprint(f"[red]Error: {error}[/red]")
    
    def handle_silent(self, error: Exception, context: str = "") -> None:
        """Handle an error silently (only log in debug mode)."""
        if self.debug:
            self.handle(error, context)
    
    def warn(self, message: str) -> None:
        """Display a warning message."""
        if self.verbose or self.debug:
            rprint(f"[yellow]{message}[/yellow]")
    
    def info(self, message: str) -> None:
        """Display an info message (debug only)."""
        if self.debug:
            rprint(f"[cyan]{message}[/cyan]")



"""
This module encapsulates all interactions with the vector database (ChromaDB).

It uses a lightweight ONNX-based model for embeddings to avoid large dependencies
like PyTorch.
"""
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import List, Dict, Any

import chromadb
# Use the lightweight ONNX embedding function included with chromadb
from chromadb.utils.embedding_functions import ONNXMiniLM_L6_V2

from aye.model.models import VectorIndexResult
from aye.model.ast_chunker import ast_chunker, get_language_from_file_path


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(os.devnull, 'w') as fnull:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = fnull
        sys.stderr = fnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


def initialize_index(root_path: Path) -> Any:
    """
    Initializes a persistent ChromaDB client and gets or creates a collection.

    Args:
        root_path: The root directory of the project.

    Returns:
        The ChromaDB collection object, ready for use.
    """
    db_path = root_path / ".aye" / "chroma_db"
    db_path.mkdir(parents=True, exist_ok=True)

    client = chromadb.PersistentClient(path=str(db_path))
    
    # Instantiate the lightweight ONNX embedding function.
    # This avoids pulling in PyTorch and is much smaller.
    # The first time this is called, it will download the ONNX model.
    # We suppress stdout/stderr to hide the download progress from the user.
    with suppress_stdout_stderr():
        embedding_function = ONNXMiniLM_L6_V2()

    # A collection is like a table in a traditional database
    collection = client.get_or_create_collection(
        name="project_code_index",
        embedding_function=embedding_function,
        metadata={"hnsw:space": "cosine"}  # Cosine similarity is good for text
    )
    return collection


def _chunk_file(content: str, chunk_size: int = 100, overlap: int = 10) -> List[str]:
    """
    Simple text chunker based on lines of code.

    TODO: This can be replaced with a more sophisticated, language-aware chunker.
    """
    lines = content.splitlines()
    if not lines:
        return []

    chunks = []
    for i in range(0, len(lines), chunk_size - overlap):
        chunk = "\n".join(lines[i:i + chunk_size])
        chunks.append(chunk)
    return chunks


def update_index_coarse(
    collection: Any, 
    files_to_update: Dict[str, str]
) -> None:
    """
    Performs a coarse, file-per-chunk update to the index.
    The document ID is the file path itself. This is for the fast, initial pass.
    """
    if not files_to_update:
        return

    ids = list(files_to_update.keys())
    documents = list(files_to_update.values())
    metadatas = [{"file_path": fp} for fp in ids]

    collection.upsert(
        ids=ids,
        documents=documents,
        metadatas=metadatas
    )


def refine_file_in_index(collection: Any, file_path: str, content: str):
    """
    Refines the index for a single file by replacing its coarse chunk
    with fine-grained, AST-based chunks.
    """
    # 1. Delete the old coarse chunk, which used the file_path as its ID.
    collection.delete(ids=[file_path])

    # 2. Create and upsert the new fine-grained chunks.
    language_name = get_language_from_file_path(file_path)
    chunks = []
    if language_name:
        chunks = ast_chunker(content, language_name)

    # Fallback to line-based chunking if AST chunking fails or is not supported
    if not chunks:
        chunks = _chunk_file(content)

    if not chunks:
        return

    ids = [f"{file_path}:{i}" for i, _ in enumerate(chunks)]
    metadatas = [{"file_path": file_path} for _ in chunks]

    collection.upsert(
        documents=chunks,
        metadatas=metadatas,
        ids=ids
    )


def delete_from_index(collection: Any, deleted_files: List[str]):
    """
    Deletes all chunks associated with a list of file paths from the index.
    This handles both coarse (id=file_path) and fine-grained chunks.
    """
    if not deleted_files:
        return
    collection.delete(where={"file_path": {"$in": deleted_files}})


def query_index(
    collection: Any, 
    query_text: str, 
    n_results: int = 10,
    min_relevance: float = 0.0
) -> List[VectorIndexResult]:
    """
    Queries the vector index for context relevant to the user's prompt.

    Args:
        collection: The ChromaDB collection object.
        query_text: The user's prompt.
        n_results: The number of relevant chunks to retrieve.
        min_relevance: The minimum similarity score (0-1) for a chunk to be considered relevant.

    Returns:
        A list of VectorIndexResult objects. If filtering by `min_relevance` yields no results,
        it falls back to returning the top 10 most relevant chunks.
    """
    if not query_text:
        return []

    results = collection.query(
        query_texts=[query_text],
        n_results=n_results,
        include=["metadatas", "documents", "distances"]
    )

    all_results = []
    # The result is a dictionary with lists for each requested field.
    # We need to iterate through them by index.
    ids = results.get('ids', [[]])[0]
    documents = results.get('documents', [[]])[0]
    metadatas = results.get('metadatas', [[]])[0]
    distances = results.get('distances', [[]])[0]

    for i in range(len(ids)):
        all_results.append(
            VectorIndexResult(
                file_path=metadatas[i].get("file_path", "unknown"),
                content=documents[i],
                score=1 - distances[i]  # Convert distance to similarity score
            )
        )
    
    if min_relevance > 0.0:
        filtered_results = [r for r in all_results if r.score >= min_relevance]

        # If filtering removes all results, but there were some initial matches,
        # fall back to returning the top 10 results regardless of score.
        # This ensures some context is always provided if anything is found at all.
        if not filtered_results and all_results:
            return all_results[:10]
        
        return filtered_results

    return all_results



import os
import json
from typing import Dict, Any, Optional
import httpx
from pathlib import Path

from rich import print as rprint

from .plugin_base import Plugin
from aye.model.config import SYSTEM_PROMPT
from aye.controller.util import is_truncated_json

LLM_TIMEOUT = 600.0
LLM_OUTPUT_TOKENS = 49152


# Message shown when LLM response is truncated due to output token limits
TRUNCATED_RESPONSE_MESSAGE = (
    "It looks like my response was cut off because it exceeded the output limit. "
    "This usually happens when you ask me to generate or modify many files at once.\n\n"
    "**To fix this, please try:**\n"
    "1. Break your request into smaller parts (e.g., one file at a time)\n"
    "2. Use the `with` command to focus on specific files: `with file1.py, file2.py: your request`\n"
    "3. Ask me to work on fewer files or smaller changes in each request\n\n"
    "For example, instead of 'update all files to add logging', try:\n"
    "  `with src/main.py: add logging to this file`"
)


class LocalModelPlugin(Plugin):
    name = "local_model"
    version = "1.0.0"
    premium = "free"

    def __init__(self):
        super().__init__()
        self.chat_history: Dict[str, list] = {}
        self.history_file: Optional[Path] = None

    def init(self, cfg: Dict[str, Any]) -> None:
        """Initialize the local model plugin."""
        super().init(cfg)
        if self.debug:
            rprint(f"[bold yellow]Initializing {self.name} v{self.version}[/]")
        
    def _load_history(self) -> None:
        """Load chat history from disk."""
        if not self.history_file:
            if self.verbose:
                rprint("[yellow]History file path not set for local model. Skipping load.[/]")
            self.chat_history = {}
            return

        if self.history_file.exists():
            try:
                data = json.loads(self.history_file.read_text(encoding="utf-8"))
                self.chat_history = data.get("conversations", {})
            except Exception as e:
                if self.verbose:
                    rprint(f"[yellow]Could not load chat history: {e}[/]")
                self.chat_history = {}
        else:
            self.chat_history = {}

    def _save_history(self) -> None:
        """Save chat history to disk."""
        if not self.history_file:
            if self.verbose:
                rprint("[yellow]History file path not set for local model. Skipping save.[/]")
            return

        try:
            self.history_file.parent.mkdir(parents=True, exist_ok=True)
            data = {"conversations": self.chat_history}
            self.history_file.write_text(json.dumps(data, indent=2), encoding="utf-8")
        except Exception as e:
            if self.verbose:
                rprint(f"[yellow]Could not save chat history: {e}[/]")

    def _get_conversation_id(self, chat_id: Optional[int] = None) -> str:
        """Get conversation ID for history tracking."""
        return str(chat_id) if chat_id and chat_id > 0 else "default"

    def _build_user_message(self, prompt: str, source_files: Dict[str, str]) -> str:
        """Build the user message with optional source files appended."""
        user_message = prompt
        if source_files:
            user_message += "\n\n--- Source files are below. ---\n"
            for file_name, content in source_files.items():
                user_message += f"\n** {file_name} **\n```\n{content}\n```\n"
        return user_message

    def _parse_llm_response(self, generated_text: str) -> Dict[str, Any]:
        """Parse LLM response text and convert to expected format."""
        try:
            llm_response = json.loads(generated_text)
        except json.JSONDecodeError:
            # Check if this looks like a truncated response
            #if is_truncated_json(generated_text):
            #    return {
            #        "summary": TRUNCATED_RESPONSE_MESSAGE,
            #        "updated_files": []
            #    }
            
            # Not truncated, just malformed - return as plain text
            llm_response = {
                "answer_summary": generated_text,
                "source_files": []
            }
        
        return {
            "summary": llm_response.get("answer_summary", ""),
            "updated_files": [
                {
                    "file_name": f.get("file_name"),
                    "file_content": f.get("file_content")
                }
                for f in llm_response.get("source_files", [])
            ]
        }

    def _create_error_response(self, error_msg: str) -> Dict[str, Any]:
        """Create a standardized error response."""
        if self.verbose:
            rprint(f"[red]{error_msg}[/]")
        return {
            "summary": error_msg,
            "updated_files": []
        }

    def _handle_databricks(self, prompt: str, source_files: Dict[str, str], chat_id: Optional[int] = None, system_prompt: Optional[str] = None) -> Optional[Dict[str, Any]]:
        api_url = os.environ.get("AYE_DBX_API_URL")
        api_key = os.environ.get("AYE_DBX_API_KEY")
        model_name = os.environ.get("AYE_DBX_MODEL", "gpt-3.5-turbo")
        
        if not api_url or not api_key:
            return None
        
        conv_id = self._get_conversation_id(chat_id)
        if conv_id not in self.chat_history:
            self.chat_history[conv_id] = []
        
        user_message = self._build_user_message(prompt, source_files)
        effective_system_prompt = system_prompt if system_prompt else SYSTEM_PROMPT
        
        messages = [{"role": "system", "content": effective_system_prompt}] + self.chat_history[conv_id] + [{"role": "user", "content": user_message}]
        
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {api_key}"}
        payload = {"model": model_name, "messages": messages, "temperature": 0.7, "max_tokens": 16384, "response_format": {"type": "json_object"}}
        
        try:
            with httpx.Client(timeout=LLM_TIMEOUT) as client:
                response = client.post(api_url, json=payload, headers=headers)
                response.raise_for_status()
                result = response.json()
                if result.get("choices") and result["choices"][0].get("message"):
                    generated_text = result["choices"][0]["message"]["content"][0]["text"]
                    self.chat_history[conv_id].append({"role": "user", "content": user_message})
                    self.chat_history[conv_id].append({"role": "assistant", "content": generated_text})
                    self._save_history()
                    return self._parse_llm_response(generated_text)
                return self._create_error_response("Failed to get a valid response from the Databricks API")
        except httpx.HTTPStatusError as e:
            error_msg = f"DBX API error: {e.response.status_code}"
            try:
                error_detail = e.response.json()
                if "error" in error_detail:
                    error_msg += f" - {error_detail['error'].get('message', str(error_detail['error']))}"
            except: error_msg += f" - {e.response.text[:200]}"
            return self._create_error_response(error_msg)
        except Exception as e:
            return self._create_error_response(f"Error calling Databricks API: {str(e)}")

    def _handle_openai_compatible(self, prompt: str, source_files: Dict[str, str], chat_id: Optional[int] = None, system_prompt: Optional[str] = None) -> Optional[Dict[str, Any]]:
        api_url = os.environ.get("AYE_LLM_API_URL")
        api_key = os.environ.get("AYE_LLM_API_KEY")
        model_name = os.environ.get("AYE_LLM_MODEL", "gpt-3.5-turbo")
        
        if not api_url or not api_key:
            return None
        
        conv_id = self._get_conversation_id(chat_id)
        if conv_id not in self.chat_history:
            self.chat_history[conv_id] = []
        
        user_message = self._build_user_message(prompt, source_files)
        effective_system_prompt = system_prompt if system_prompt else SYSTEM_PROMPT
        messages = [{"role": "system", "content": effective_system_prompt}] + self.chat_history[conv_id] + [{"role": "user", "content": user_message}]
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {api_key}"}
        payload = {"model": model_name, "messages": messages, "temperature": 0.7, "max_tokens": LLM_OUTPUT_TOKENS, "response_format": {"type": "json_object"}}
        
        try:
            with httpx.Client(timeout=LLM_TIMEOUT) as client:
                response = client.post(api_url, json=payload, headers=headers)
                response.raise_for_status()
                result = response.json()
                if result.get("choices") and result["choices"][0].get("message"):
                    generated_text = result["choices"][0]["message"]["content"]
                    self.chat_history[conv_id].append({"role": "user", "content": user_message})
                    self.chat_history[conv_id].append({"role": "assistant", "content": generated_text})
                    self._save_history()
                    return self._parse_llm_response(generated_text)
                return self._create_error_response("Failed to get a valid response from the OpenAI-compatible API")
        except httpx.HTTPStatusError as e:
            error_msg = f"OpenAI API error: {e.response.status_code}"
            try:
                error_detail = e.response.json()
                if "error" in error_detail:
                    error_msg += f" - {error_detail['error'].get('message', str(error_detail['error']))}"
            except: error_msg += f" - {e.response.text[:200]}"
            return self._create_error_response(error_msg)
        except Exception as e:
            return self._create_error_response(f"Error calling OpenAI-compatible API: {str(e)}")

    def _handle_gemini_pro_25(self, prompt: str, source_files: Dict[str, str], chat_id: Optional[int] = None, system_prompt: Optional[str] = None) -> Optional[Dict[str, Any]]:
        api_key = os.environ.get("GEMINI_API_KEY")
        if not api_key:
            return None

        conv_id = self._get_conversation_id(chat_id)
        if conv_id not in self.chat_history:
            self.chat_history[conv_id] = []

        user_message = self._build_user_message(prompt, source_files)
        url = "https://generativelanguage.googleapis.com/v1beta/models/gemini-2.5-pro:generateContent"
        headers = {"Content-Type": "application/json", "x-goog-api-key": api_key}
        
        contents = [{"role": "user" if msg["role"] == "user" else "model", "parts": [{"text": msg["content"]}]} for msg in self.chat_history[conv_id]]
        contents.append({"role": "user", "parts": [{"text": user_message}]})
        
        effective_system_prompt = system_prompt if system_prompt else SYSTEM_PROMPT
        payload = {"contents": contents, "systemInstruction": {"parts": [{"text": effective_system_prompt}]}, "generationConfig": {"temperature": 0.7, "topK": 40, "topP": 0.95, "maxOutputTokens": LLM_OUTPUT_TOKENS, "responseMimeType": "application/json"}}

        try:
            with httpx.Client(timeout=LLM_TIMEOUT) as client:
                response = client.post(url, json=payload, headers=headers)
                response.raise_for_status()
                result = response.json()
                if result.get("candidates") and result["candidates"][0].get("content"):
                    generated_text = result["candidates"][0]["content"]["parts"][0].get("text", "")
                    self.chat_history[conv_id].append({"role": "user", "content": user_message})
                    self.chat_history[conv_id].append({"role": "assistant", "content": generated_text})
                    self._save_history()
                    return self._parse_llm_response(generated_text)
                return self._create_error_response("Failed to get a valid response from Gemini API")
        except httpx.HTTPStatusError as e:
            return self._create_error_response(f"Gemini API error: {e.response.status_code} - {e.response.text}")
        except Exception as e:
            return self._create_error_response(f"Error calling Gemini API: {str(e)}")

    def on_command(self, command_name: str, params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        if command_name == "new_chat":
            root = params.get("root")
            history_file = Path(root) / ".aye" / "chat_history.json" if root else Path.cwd() / ".aye" / "chat_history.json"
            history_file.unlink(missing_ok=True)
            self.chat_history = {}
            if self.verbose: rprint("[yellow]Local model chat history cleared.[/]")
            return {"status": "local_history_cleared"}

        if command_name == "local_model_invoke":
            prompt = params.get("prompt", "").strip()
            model_id = params.get("model_id", "")
            source_files = params.get("source_files", {})
            chat_id = params.get("chat_id")
            root = params.get("root")
            system_prompt = params.get("system_prompt")

            self.history_file = Path(root) / ".aye" / "chat_history.json" if root else Path.cwd() / ".aye" / "chat_history.json"
            self._load_history()

            result = self._handle_openai_compatible(prompt, source_files, chat_id, system_prompt)
            if result is not None: return result

            result = self._handle_databricks(prompt, source_files, chat_id, system_prompt)
            if result is not None: return result

            if model_id == "google/gemini-2.5-pro":
                return self._handle_gemini_pro_25(prompt, source_files, chat_id, system_prompt)
            
            return None

        return None



# Test suite for aye.model.version_checker module
import io
import sys
from unittest import TestCase
from unittest.mock import patch, MagicMock

from aye.model import version_checker


class TestVersionChecker(TestCase):
    """Test suite for version checking functionality."""

    def test_get_current_version_success(self):
        """Test getting current version when package is installed."""
        with patch("importlib.metadata.version") as mock_version:
            mock_version.return_value = "0.26.0"
            result = version_checker.get_current_version()
            self.assertEqual(result, "0.26.0")

    def test_get_current_version_package_not_found(self):
        """Test fallback when package is not found."""
        from importlib.metadata import PackageNotFoundError
        with patch("importlib.metadata.version") as mock_version:
            mock_version.side_effect = PackageNotFoundError()
            result = version_checker.get_current_version()
            self.assertEqual(result, "0.0.0")

    def test_get_latest_stable_version_info_success(self):
        """Test fetching latest stable version from PyPI successfully."""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "info": {
                "version": "0.27.0",
                "requires_python": ">=3.8, <3.14"
            },
            "releases": {
                "0.25.0": [],
                "0.26.0": [],
                "0.27.0": [],
                "0.28.0a1": [],  # Prerelease should be excluded
            }
        }

        with patch("aye.model.version_checker.httpx.get") as mock_get:
            mock_get.return_value = mock_response
            result = version_checker.get_latest_stable_version_info()

            self.assertIsNotNone(result)
            version_str, python_requires = result
            self.assertEqual(version_str, "0.27.0")
            self.assertEqual(python_requires, ">=3.8, <3.14")

    def test_get_latest_stable_version_info_excludes_prereleases(self):
        """Test that prereleases are excluded from version selection."""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "info": {
                "version": "0.28.0a1",
                "requires_python": ">=3.8, <3.14"
            },
            "releases": {
                "0.25.0": [],
                "0.26.0": [],
                "0.27.0": [],
                "0.28.0a1": [],  # Latest is prerelease
                "0.28.0b1": [],  # Beta
                "0.28.0rc1": [],  # Release candidate
            }
        }

        with patch("aye.model.version_checker.httpx.get") as mock_get:
            mock_get.return_value = mock_response
            result = version_checker.get_latest_stable_version_info()

            self.assertIsNotNone(result)
            version_str, python_requires = result
            # Should return 0.27.0, not the 0.28.0 prereleases
            self.assertEqual(version_str, "0.27.0")

    def test_get_latest_stable_version_info_network_failure(self):
        """Test graceful failure when network request fails."""
        with patch("aye.model.version_checker.httpx.get") as mock_get:
            mock_get.side_effect = Exception("Network error")
            result = version_checker.get_latest_stable_version_info()
            self.assertIsNone(result)

    def test_get_latest_stable_version_info_timeout(self):
        """Test graceful failure on timeout."""
        with patch("aye.model.version_checker.httpx.get") as mock_get:
            import httpx
            mock_get.side_effect = httpx.TimeoutException("Timeout")
            result = version_checker.get_latest_stable_version_info()
            self.assertIsNone(result)

    def test_is_newer_version_available_true(self):
        """Test detection when newer version is available."""
        with patch("aye.model.version_checker.get_current_version") as mock_current, \
             patch("aye.model.version_checker.get_latest_stable_version_info") as mock_latest:
            mock_current.return_value = "0.26.0"
            mock_latest.return_value = ("0.27.0", ">=3.8, <3.14")

            is_newer, latest, current, python_requires = version_checker.is_newer_version_available()

            self.assertTrue(is_newer)
            self.assertEqual(latest, "0.27.0")
            self.assertEqual(current, "0.26.0")
            self.assertEqual(python_requires, ">=3.8, <3.14")

    def test_is_newer_version_available_false_same_version(self):
        """Test when current version equals latest."""
        with patch("aye.model.version_checker.get_current_version") as mock_current, \
             patch("aye.model.version_checker.get_latest_stable_version_info") as mock_latest:
            mock_current.return_value = "0.26.0"
            mock_latest.return_value = ("0.26.0", ">=3.8, <3.14")

            is_newer, latest, current, python_requires = version_checker.is_newer_version_available()

            self.assertFalse(is_newer)
            self.assertEqual(latest, "0.26.0")
            self.assertEqual(current, "0.26.0")

    def test_is_newer_version_available_false_older_version(self):
        """Test when current version is newer than PyPI (dev version)."""
        with patch("aye.model.version_checker.get_current_version") as mock_current, \
             patch("aye.model.version_checker.get_latest_stable_version_info") as mock_latest:
            mock_current.return_value = "0.28.0"
            mock_latest.return_value = ("0.27.0", ">=3.8, <3.14")

            is_newer, latest, current, python_requires = version_checker.is_newer_version_available()

            self.assertFalse(is_newer)
            self.assertEqual(latest, "0.27.0")
            self.assertEqual(current, "0.28.0")

    def test_is_newer_version_available_network_failure(self):
        """Test when latest version cannot be fetched."""
        with patch("aye.model.version_checker.get_current_version") as mock_current, \
             patch("aye.model.version_checker.get_latest_stable_version_info") as mock_latest:
            mock_current.return_value = "0.26.0"
            mock_latest.return_value = None

            is_newer, latest, current, python_requires = version_checker.is_newer_version_available()

            self.assertFalse(is_newer)
            self.assertIsNone(latest)
            self.assertEqual(current, "0.26.0")
            self.assertIsNone(python_requires)

    def test_is_newer_version_available_excludes_prerelease(self):
        """Test that prereleases are not considered as newer versions."""
        with patch("aye.model.version_checker.get_current_version") as mock_current, \
             patch("aye.model.version_checker.get_latest_stable_version_info") as mock_latest:
            mock_current.return_value = "0.26.0"
            # Latest stable is 0.26.0, even though 0.27.0a1 exists
            mock_latest.return_value = ("0.26.0", ">=3.8, <3.14")

            is_newer, latest, current, python_requires = version_checker.is_newer_version_available()

            # Should not report an update since prereleases are excluded
            self.assertFalse(is_newer)
            self.assertEqual(latest, "0.26.0")
            self.assertEqual(current, "0.26.0")

    def test_parse_python_version_max_exclusive_upper_bound(self):
        """Test parsing exclusive upper bound like <3.14."""
        result = version_checker._parse_python_version_max(">=3.8, <3.14")
        self.assertEqual(result, "3.13")

    def test_parse_python_version_max_inclusive_upper_bound(self):
        """Test parsing inclusive upper bound like <=3.13."""
        result = version_checker._parse_python_version_max(">=3.8, <=3.13")
        self.assertEqual(result, "3.13")

    def test_parse_python_version_max_no_upper_bound(self):
        """Test when there's no upper bound specified."""
        result = version_checker._parse_python_version_max(">=3.8")
        self.assertIsNone(result)

    def test_parse_python_version_max_none_input(self):
        """Test with None input."""
        result = version_checker._parse_python_version_max(None)
        self.assertIsNone(result)

    def test_check_version_and_print_warning_with_update(self):
        """Test that warning is printed when update is available."""
        with patch("aye.model.version_checker.is_newer_version_available") as mock_check:
            mock_check.return_value = (True, "0.27.0", "0.26.0", ">=3.8, <3.14")

            # Capture stdout
            captured_output = io.StringIO()
            sys.stdout = captured_output

            try:
                version_checker.check_version_and_print_warning()
                output = captured_output.getvalue()

                # Verify warning message components
                self.assertIn("Notice", output)
                self.assertIn("newer version", output)
                self.assertIn("0.26.0", output)
                self.assertIn("0.27.0", output)
                self.assertIn("3.13", output)  # Max Python version
                self.assertIn("pip install --upgrade ayechat", output)
            finally:
                sys.stdout = sys.__stdout__

    def test_check_version_and_print_warning_with_update_no_python_info(self):
        """Test warning without Python version information."""
        with patch("aye.model.version_checker.is_newer_version_available") as mock_check:
            mock_check.return_value = (True, "0.27.0", "0.26.0", None)

            # Capture stdout
            captured_output = io.StringIO()
            sys.stdout = captured_output

            try:
                version_checker.check_version_and_print_warning()
                output = captured_output.getvalue()

                # Verify warning message components
                self.assertIn("Notice", output)
                self.assertIn("0.26.0", output)
                self.assertIn("0.27.0", output)
                # Should not contain Python version info
                self.assertNotIn("Supports Python", output)
            finally:
                sys.stdout = sys.__stdout__

    def test_check_version_and_print_warning_no_update(self):
        """Test that no warning is printed when no update is available."""
        with patch("aye.model.version_checker.is_newer_version_available") as mock_check:
            mock_check.return_value = (False, "0.26.0", "0.26.0", ">=3.8, <3.14")

            # Capture stdout
            captured_output = io.StringIO()
            sys.stdout = captured_output

            try:
                version_checker.check_version_and_print_warning()
                output = captured_output.getvalue()

                # Should be empty or minimal output
                self.assertEqual(output, "")
            finally:
                sys.stdout = sys.__stdout__

    def test_check_version_and_print_warning_network_failure(self):
        """Test that no error is shown when version check fails."""
        with patch("aye.model.version_checker.is_newer_version_available") as mock_check:
            mock_check.return_value = (False, None, "0.26.0", None)

            # Capture stdout
            captured_output = io.StringIO()
            sys.stdout = captured_output

            try:
                version_checker.check_version_and_print_warning()
                output = captured_output.getvalue()

                # Should not print anything on network failure
                self.assertEqual(output, "")
            finally:
                sys.stdout = sys.__stdout__
